(* This sample client performs type inference for a fragment of ML and translates
   it down to a fragment of System F. *)

(* -------------------------------------------------------------------------- *)

(* During inference, we will need an environment to keep trace of
 * the user defined types *)

module Env = struct
  module LabelId = struct
    type t = ML.label_id
    let compare = Stdlib.compare
  end
                 
  module LabelIdMap =
    Map.Make(LabelId)

  type label2descr = ML.label_descr LabelIdMap.t

  module TypeId = struct
    type t = ML.type_id
    let compare = Stdlib.compare
  end

  module TypeIdMap =
    Map.Make(TypeId)

  type data_type2descr = ML.data_type_descr TypeIdMap.t

  (* Environment for user defined data types *)
  type t = {
      data_types : data_type2descr ;
      labels : label2descr
    }

  exception UnexpectedRecord

  let empty = {
      data_types = TypeIdMap.empty;
      labels = LabelIdMap.empty;
    }

  let add_data_type m tdescr =
    TypeIdMap.add ML.(tdescr.name) tdescr m

  let add_labels m ldescrs =
    List.fold_left (fun m ldescr ->
        LabelIdMap.add ML.(ldescr.label_name) ldescr m
      ) m ldescrs

  let add_typedecl e tdescr = {
      data_types = add_data_type e.data_types tdescr;
      labels = add_labels e.labels ML.(tdescr.labels_descr);
    }

  let descr_from_label { labels ; _ } (ML.Label l) =
    LabelIdMap.find (ML.Label l) labels

  let type_descr_from_type { data_types ; _ } (ML.Type tid) =
    TypeIdMap.find (ML.Type tid) data_types

  let label_index (ML.({ labels_descr ; _ })) l =
    let combine i ldescr = (i, ML.(ldescr.label_name)) in
    List.mapi combine labels_descr
    |> List.find (fun (_, l') -> l = l')
    |> fst
end

(* -------------------------------------------------------------------------- *)

(* Translation from ML types to F types *)

let fresh_tyvar =
  let cpt = ref 0 in
  
  (* We generate odd number for the user and even number for the solver *)
  fun () -> incr cpt; 2 * !cpt + 1

(* Here [rec_occur] is the function called when we encounter an
 * occurence of the name of the type we translate *)
let rec translate_type
  (env : Env.t)
  (params : (string * F.nominal_type) list)
  (rec_occur : unit -> F.nominal_type)
  (name_self : ML.type_id)
  (ty : ML.nominal_type)
  : F.nominal_type =
  let rec tt (ty : ML.nominal_type) : F.nominal_type =
    match ty with
    | ML.TyVar x ->
       List.assoc x params
      
    | ML.TyArrow (ty1, ty2) ->
       F.TyArrow (tt ty1, tt ty2)
      
    | ML.TyProduct tys ->
       F.TyProduct (List.map tt tys)
      
    | ML.TyConstr (tid, tys) ->
       if tid = name_self then begin
           let desc = Env.type_descr_from_type env name_self in

           (* This assert should hold as long as we only treat
            * regular recursion, that is types in which each type parameters
            * is instantiated with only one type *)
           assert (tys = List.map (fun x -> ML.TyVar x) desc.ML.type_params);

           rec_occur ()
         end
       else
         translate_type_expr env (List.map tt tys) tid
  in
  tt ty

and translate_typedecl
  (env : Env.t)
  (params : F.nominal_type list)
  (desc : ML.data_type_descr)
  (rec_occur : unit -> F.nominal_type)
    : F.nominal_type list =
  let params = List.combine ML.(desc.type_params) params in
  let tys =
    List.map (fun ldescr -> ML.(ldescr.arg_type)) ML.(desc.labels_descr)
    |> List.map (translate_type env params rec_occur ML.(desc.name))
  in
  tys

and translate_type_expr
  (env : Env.t)
  (params : F.nominal_type list)
  (type_id : ML.type_id)
  : F.nominal_type =
  let desc = Env.type_descr_from_type env type_id in
  let tau = fresh_tyvar () in
  let rec_occur () = F.TyVar tau in
  let tys = translate_typedecl env params desc rec_occur in
  F.TyMu (tau, F.TySum tys)

let translate_constr_types
  (env : Env.t)
  (params : F.nominal_type list)
  (type_id : ML.type_id)
  : F.nominal_type list =
  let desc = Env.type_descr_from_type env type_id in
  let rec_occur () =
    translate_type_expr env params type_id
  in
  translate_typedecl env params desc rec_occur

(* -------------------------------------------------------------------------- *)

(* The unifier will use the following type structure. *)

module S = struct

  type 'a structure =
    | TyArrow of 'a * 'a
    | TyProduct of 'a list
    | TyConstr of ML.type_id * Env.t * 'a list

  let map f t =
    match t with
    | TyArrow (t1, t2) ->
        let t1 = f t1 in
        let t2 = f t2 in
        TyArrow (t1, t2)
    | TyProduct ts ->
        let ts = List.map f ts in
        TyProduct ts
    | TyConstr (tid, env, ts) ->
       let ts = List.map f ts in
       TyConstr (tid, env, ts)

  let fold f t accu =
    match t with
    | TyArrow (t1, t2) ->
        let accu = f t1 accu in
        let accu = f t2 accu in
        accu
    | TyProduct ts ->
        List.fold_right f ts accu
    | TyConstr (_, _, ts) ->
        List.fold_right f ts accu

  let iter f t =
    let _ = map f t in
    ()

  exception Iter2

  let iter2 f t u =
    match t, u with
    | TyArrow (t1, t2), TyArrow (u1, u2) ->
        f t1 u1;
        f t2 u2
    | TyProduct ts1, TyProduct ts2
    | TyConstr (_, _, ts1), TyConstr (_, _, ts2) ->
        if List.length ts1 <> List.length ts2 then raise Iter2;
        List.iter2 f ts1 ts2
    | TyArrow _, _
    | TyProduct _, _
    | TyConstr _, _ ->
        raise Iter2

end

(* -------------------------------------------------------------------------- *)

(* The unifier type structure is decoded into the target calculus type structure
   as follows. *)

module O = struct

  type tyvar =
    int

  (* See also [fresh_tyvar] *)        
  let solver_tyvar n =
    2 * n

  type 'a structure =
    'a S.structure

  type ty =
    F.nominal_type

  let variable x =
    F.TyVar x

  let structure t =
    match t with
    | S.TyArrow (t1, t2) ->
        F.TyArrow (t1, t2)
    | S.TyProduct ts ->
        F.TyProduct ts
    | S.TyConstr (tid, env, ts) ->
        translate_type_expr env ts tid
      
  let mu x t =
    F.TyMu (x, t)

  type scheme =
    tyvar list * ty

end

(* -------------------------------------------------------------------------- *)

(* Instantiate the solver. *)

module Solver =
  Inferno.SolverHi.Make(struct include String type tevar = t end)(S)(O)

open Solver

(* -------------------------------------------------------------------------- *)

let arrow x y =
  S.TyArrow (x, y)

let product xs =
  S.TyProduct xs

let constr c env xs =
  S.TyConstr (c, env, xs)

(* Should we use smart constructors to eliminate redundant coercions when possible? *)
let smart =
  true

let flet (x, t, u) =
  match t with
  | F.Var y when smart && x = y ->
      u
  | t ->
      F.Let (x, t, u)

(* -------------------------------------------------------------------------- *)

(* The coercion [coerce vs1 vs2] converts a type of the form [forall vs1, _]
   to a type of the form [forall vs2, _], where [vs2] forms a subset of [vs1].
   This coercion allows getting rid of unused quantifiers and/or re-ordering
   quantifiers. *)

type coercion =
  F.nominal_term -> F.nominal_term

let bottom : F.nominal_type =
  let a : F.tyvar = 0 (* arbitrary *) in
  F.TyForall (a, F.TyVar a)

(* [ftyabs1 v t] builds a (capital-Lambda) abstraction of the type variable [v]
   in the term [t]. It is a smart constructor: if it recognizes an eta-redex,
   it contracts it on the fly. We are in a special case where, if [v] and [w]
   are the same variable, then this variable does not occur free in [t], so we
   don't need to perform this costly check at runtime. This eta-contraction is
   not essential anyway; it's just a way of avoiding coercion clutter in the
   common case where the coercion actually has no effect. *)

let ftyabs1 v t =
  match t with
  | F.TyApp (t, F.TyVar w) when smart && v = w ->
      t
  | t ->
      F.TyAbs (v, t)

(* TEMPORARY find a better name for [coerce] *)

let coerce (vs1 : O.tyvar list) (vs2 : O.tyvar list) : coercion =
  (* Assume the term [t] has type [forall vs1, _]. *)
  fun t ->
    (* Introduce the desired quantifiers. *)
    List.fold_right ftyabs1 vs2 (
      (* Now, specialize the term [t]. For each member of [vs1],
         we must provide a suitable instantiation. *)
      F.ftyapp t (
        (* [vs1] is a superset of [vs2]. For each member of [vs1], if it is a
           member of [vs2], then we keep it (by instantiating it with itself),
           otherwise we get rid of it (by instantiating it with an arbitrary
           closed type, say [bottom]). *)
        let suitable (v : O.tyvar) : O.ty =
          if List.mem v vs2 then F.TyVar v else bottom
          (* TEMPORARY need an efficient membership test in [vs2] *)
        in
        List.map suitable vs1
      )
    )

(* -------------------------------------------------------------------------- *)  

(* Might want to move this function into SolverHi.ml *)
let rec bind_list (f : ('a -> ('b, 'r) binder)) (xs : 'a list)
  : ('b list, 'r) binder
  = fun k ->
  match xs with
  | [] ->
     k []
  | x :: xs ->
     let@ y = f x in
     let@ ys = bind_list f xs in
     k (y :: ys)

let rec bind2_list (f : ('a -> ('b * 'c co, 'r) binder)) (xs : 'a list)
  : ('b list * 'c list co, 'r) binder
  = fun k ->
  match xs with
  | [] ->
     k ([], pure [])
  | x :: xs ->
     let@ (y, z) = f x in
     let@ (ys, zs) = bind2_list f xs in
     k (y::ys,
        let+z = z
        and+ zs = zs
        in z :: zs
       )

let instantiate_params (params : string list)
  : ((string * variable) list * (string * O.ty) list co, 'r) binder =
  let on_param (x : string) =
    fun k ->
    let@ (y, ty) = exist in
    k ((x, y), (pure x ^& ty))
  in
  bind2_list on_param params

let rec convert (env : Env.t) (params : (string * variable) list) (ty : ML.nominal_type)
  : (variable, 'r) binder
  = fun k ->
    match ty with
    | ML.TyVar tx ->
       let w = List.assoc tx params in
       k w
       
    | ML.TyArrow (ty1, ty2) ->
       let@ w1 = convert env params ty1 in
       let@ w2 = convert env params ty2 in
       let@ w = exist_ in
       let+ () = w --- arrow w1 w2
       and+ res = k w                
       in res

    | ML.TyProduct tys ->
       let@ w = exist_ in
       let@ ws = bind_list (convert env params) tys in
       let+ () = w --- product ws
       and+ res = k w
       in res

    | ML.TyConstr (tid, tys) ->
       let@ w = exist_ in
       let@ ws = bind_list (convert env params) tys in
       let+ () = w --- constr tid env ws
       and+ res = k w
       in res
  
(* -------------------------------------------------------------------------- *)

(* The client uses the combinators provided by the solver so as to transparently
   1- analyse the source term and produce constraints; and 2- decode the solution
   of the constraints and produce a term in the target calculus. These two steps
   take place in different phases, but the code is written as if there was just
   one phase. *)

(* The function [analyse] takes a source term [t] and an expected type [w].
   No type environment is required, as everything is built into the constraint via
   suitable combinators, such as [def]. *)

(* BEGIN HASTYPE *)
let hastype (e : Env.t) (t : ML.term) (w : variable) : F.nominal_term co
  =
  let rec hastype t w =
    match t with

     (* Variable. *)
    | ML.Var x ->
     
       (* [w] must be an instance of the type scheme associated with [x]. *)
       let+ tys = instance x w in
       (* The translation makes the type application explicit. *)
       F.ftyapp (F.Var x) tys
       
     (* Abstraction. *)
    | ML.Abs (x, u) ->

       (* We do not know a priori what the domain and codomain of this function
          are, so we must infer them. We introduce two type variables to stand
          for these unknowns. *)
       let@ (v1, ty1) = exist in
       (* Here, we could use [exist_], because we do not need [ty2]. I refrain
          from using it, just to simplify the paper. *)
       let@ (v2, _ty2) = exist in
       (* [w] must be the function type [v1 -> v2]. *)
       let+ () = w --- arrow v1 v2
       (* Under the assumption that [x] has type [domain], the term [u] must
          have type [codomain]. *)
       and+ u' = def x v1 (hastype u v2)
       and+ ty1 = ty1
       in           
       (* Once these constraints are solved, we obtain the translated function
         body [u']. There remains to construct an explicitly-typed abstraction
         in the target calculus. *)
       F.Abs (x, ty1, u')
    
    | ML.App (t1, t2) ->
       
       (* Introduce a type variable to stand for the unknown argument type. *)
       let@ (v, _ty) = exist in
       let+ t1' = lift hastype t1 (arrow v w)
       and+ t2' = hastype t2 v
       in F.App (t1', t2')

     (* Generalization. *)
    | ML.Let (x, t, u) ->
     
       (* Construct a ``let'' constraint. *)
       let+ ((b, _), a, t', u') = let1 x (hastype t) (hastype u w) in
       (* [a] are the type variables that we must introduce (via Lambda-abstractions)
         while type-checking [t]. [(b, _)] is the type scheme that [x] must receive
         while type-checking [u]. Its quantifiers [b] are guaranteed to form a subset of
         [a]. Hence, in general, we must re-bind [x] to an application of a suitable
         coercion to [x]. We use smart constructors so that, if the lists [a] and
         [b] happen to be equal, no extra code is produced. *)
       F.Let (x, F.ftyabs a t',
(* IFPAPER
       F.Let (x, coerce a b (F.Var x),
       ELSE *)
       flet (x, coerce a b (F.Var x),
(* END *)
       u'))
(* END HASTYPE *)

    | ML.Tuple ts ->
       let on_term (t:ML.term) : ('b * 'c co, 'r) binder =
         fun (k : ('b * 'c co) -> 'r co) : 'r co ->
         let@ v : 'b = exist_ in
         let t = hastype t v in
         k (v, t)
       in

       let@ (vs, ts') = bind2_list on_term ts in
       let+ () = w --- product vs
       and+ ts' = ts'
       in F.Tuple ts'

    | ML.LetProd (xs, t, u) ->
       let on_var (x:ML.tevar) : ('a, 'r) binder =
         fun (k : 'b -> 'r co) : 'r co ->
         let@ v = exist_ in
         def x v (k v)
       in

       let@ vs = bind_list on_var xs in
       let+ t' = lift hastype t (product vs)
       and+ u' = hastype u w
       in F.LetProd(xs, t', u')
       
    | ML.Variant (c, t) ->
       let ML.{ type_name ; arg_type ; _ } =
         Env.descr_from_label e c in
      
       let ML.{ type_params ; data_kind ; _ } as tdescr =
         Env.type_descr_from_type e type_name in
      
       if not data_kind then
         raise Env.UnexpectedRecord;

       (* We begin by instantiating the type parameters of the variant's type *)
       let@ param_vars, param_types = instantiate_params type_params in

       (* Inference variable for the argument of the variant *)
       let@ w1 = convert e param_vars arg_type in

       (* Retrieve the index of the constructor in the list of labels_descr *)
       let i = Env.label_index tdescr c in

       let+ () = w --- (constr type_name e @@ List.map snd param_vars)
       and+ param_types = param_types
       and+ t = hastype t w1
       in
       let sumtype =
         translate_constr_types e (List.map snd param_types) type_name in
       F.Inj (sumtype, i, t)
  in
  hastype t w

(* The top-level wrapper uses [let0]. It does not require an expected
   type; it creates its own using [exist]. And it runs the solver. *)

exception Unbound = Solver.Unbound
exception Unify = Solver.Unify
exception Cycle = Solver.Cycle

let translate (e : Env.t) (t : ML.term) : F.nominal_term =
  solve false (
    let0 (exist_ (hastype e t)) <$$> fun (vs, t) ->
    (* [vs] are the binders that we must introduce *)
    F.ftyabs vs t
  )
